/* win32bitmap.cpp
Implements the Bitmap class
(C) 2000 N. Douglas
13th April 2000
*/

#include "win32bitmap.h"
#include "resource.h"
#include <stdio.h>

#define START_DDOP() do {
#define END_DDOP(ddrval) } while(ddrval!=DD_OK && DDErrH(ddrval, __LINE__))

/*
 * D3DAppErrorToString
 */
static const char *D3DAppErrorToString(HRESULT error)
{
    switch(error) {
        case DD_OK:
            return "No error.\0";
        case DDERR_ALREADYINITIALIZED:
            return "This object is already initialized.\0";
        case DDERR_BLTFASTCANTCLIP:
            return "Return if a clipper object is attached to the source surface passed into a BltFast call.\0";
        case DDERR_CANNOTATTACHSURFACE:
            return "This surface can not be attached to the requested surface.\0";
        case DDERR_CANNOTDETACHSURFACE:
            return "This surface can not be detached from the requested surface.\0";
        case DDERR_CANTCREATEDC:
            return "Windows can not create any more DCs.\0";
        case DDERR_CANTDUPLICATE:
            return "Can't duplicate primary & 3D surfaces, or surfaces that are implicitly created.\0";
        case DDERR_CLIPPERISUSINGHWND:
            return "An attempt was made to set a cliplist for a clipper object that is already monitoring an hwnd.\0";
        case DDERR_COLORKEYNOTSET:
            return "No src color key specified for this operation.\0";
        case DDERR_CURRENTLYNOTAVAIL:
            return "Support is currently not available.\0";
        case DDERR_DIRECTDRAWALREADYCREATED:
            return "A DirectDraw object representing this driver has already been created for this process.\0";
        case DDERR_EXCEPTION:
            return "An exception was encountered while performing the requested operation.\0";
        case DDERR_EXCLUSIVEMODEALREADYSET:
            return "An attempt was made to set the cooperative level when it was already set to exclusive.\0";
        case DDERR_GENERIC:
            return "Generic failure.\0";
        case DDERR_HEIGHTALIGN:
            return "Height of rectangle provided is not a multiple of reqd alignment.\0";
        case DDERR_HWNDALREADYSET:
            return "The CooperativeLevel HWND has already been set. It can not be reset while the process has surfaces or palettes created.\0";
        case DDERR_HWNDSUBCLASSED:
            return "HWND used by DirectDraw CooperativeLevel has been subclassed, this prevents DirectDraw from restoring state.\0";
        case DDERR_IMPLICITLYCREATED:
            return "This surface can not be restored because it is an implicitly created surface.\0";
        case DDERR_INCOMPATIBLEPRIMARY:
            return "Unable to match primary surface creation request with existing primary surface.\0";
        case DDERR_INVALIDCAPS:
            return "One or more of the caps bits passed to the callback are incorrect.\0";
        case DDERR_INVALIDCLIPLIST:
            return "DirectDraw does not support the provided cliplist.\0";
        case DDERR_INVALIDDIRECTDRAWGUID:
            return "The GUID passed to DirectDrawCreate is not a valid DirectDraw driver identifier.\0";
        case DDERR_INVALIDMODE:
            return "DirectDraw does not support the requested mode.\0";
        case DDERR_INVALIDOBJECT:
            return "DirectDraw received a pointer that was an invalid DIRECTDRAW object.\0";
        case DDERR_INVALIDPARAMS:
            return "One or more of the parameters passed to the function are incorrect.\0";
        case DDERR_INVALIDPIXELFORMAT:
            return "The pixel format was invalid as specified.\0";
        case DDERR_INVALIDPOSITION:
            return "Returned when the position of the overlay on the destination is no longer legal for that destination.\0";
        case DDERR_INVALIDRECT:
            return "Rectangle provided was invalid.\0";
        case DDERR_LOCKEDSURFACES:
            return "Operation could not be carried out because one or more surfaces are locked.\0";
        case DDERR_NO3D:
            return "There is no 3D present.\0";
        case DDERR_NOALPHAHW:
            return "Operation could not be carried out because there is no alpha accleration hardware present or available.\0";
        case DDERR_NOBLTHW:
            return "No blitter hardware present.\0";
        case DDERR_NOCLIPLIST:
            return "No cliplist available.\0";
        case DDERR_NOCLIPPERATTACHED:
            return "No clipper object attached to surface object.\0";
        case DDERR_NOCOLORCONVHW:
            return "Operation could not be carried out because there is no color conversion hardware present or available.\0";
        case DDERR_NOCOLORKEY:
            return "Surface doesn't currently have a color key\0";
        case DDERR_NOCOLORKEYHW:
            return "Operation could not be carried out because there is no hardware support of the destination color key.\0";
        case DDERR_NOCOOPERATIVELEVELSET:
            return "Create function called without DirectDraw object method SetCooperativeLevel being called.\0";
        case DDERR_NODC:
            return "No DC was ever created for this surface.\0";
        case DDERR_NODDROPSHW:
            return "No DirectDraw ROP hardware.\0";
        case DDERR_NODIRECTDRAWHW:
            return "A hardware-only DirectDraw object creation was attempted but the driver did not support any hardware.\0";
        case DDERR_NOEMULATION:
            return "Software emulation not available.\0";
        case DDERR_NOEXCLUSIVEMODE:
            return "Operation requires the application to have exclusive mode but the application does not have exclusive mode.\0";
        case DDERR_NOFLIPHW:
            return "Flipping visible surfaces is not supported.\0";
        case DDERR_NOGDI:
            return "There is no GDI present.\0";
        case DDERR_NOHWND:
            return "Clipper notification requires an HWND or no HWND has previously been set as the CooperativeLevel HWND.\0";
        case DDERR_NOMIRRORHW:
            return "Operation could not be carried out because there is no hardware present or available.\0";
        case DDERR_NOOVERLAYDEST:
            return "Returned when GetOverlayPosition is called on an overlay that UpdateOverlay has never been called on to establish a destination.\0";
        case DDERR_NOOVERLAYHW:
            return "Operation could not be carried out because there is no overlay hardware present or available.\0";
        case DDERR_NOPALETTEATTACHED:
            return "No palette object attached to this surface.\0";
        case DDERR_NOPALETTEHW:
            return "No hardware support for 16 or 256 color palettes.\0";
        case DDERR_NORASTEROPHW:
            return "Operation could not be carried out because there is no appropriate raster op hardware present or available.\0";
        case DDERR_NOROTATIONHW:
            return "Operation could not be carried out because there is no rotation hardware present or available.\0";
        case DDERR_NOSTRETCHHW:
            return "Operation could not be carried out because there is no hardware support for stretching.\0";
        case DDERR_NOT4BITCOLOR:
            return "DirectDrawSurface is not in 4 bit color palette and the requested operation requires 4 bit color palette.\0";
        case DDERR_NOT4BITCOLORINDEX:
            return "DirectDrawSurface is not in 4 bit color index palette and the requested operation requires 4 bit color index palette.\0";
        case DDERR_NOT8BITCOLOR:
            return "DirectDrawSurface is not in 8 bit color mode and the requested operation requires 8 bit color.\0";
        case DDERR_NOTAOVERLAYSURFACE:
            return "Returned when an overlay member is called for a non-overlay surface.\0";
        case DDERR_NOTEXTUREHW:
            return "Operation could not be carried out because there is no texture mapping hardware present or available.\0";
        case DDERR_NOTFLIPPABLE:
            return "An attempt has been made to flip a surface that is not flippable.\0";
        case DDERR_NOTFOUND:
            return "Requested item was not found.\0";
        case DDERR_NOTLOCKED:
            return "Surface was not locked.  An attempt to unlock a surface that was not locked at all, or by this process, has been attempted.\0";
        case DDERR_NOTPALETTIZED:
            return "The surface being used is not a palette-based surface.\0";
        case DDERR_NOVSYNCHW:
            return "Operation could not be carried out because there is no hardware support for vertical blank synchronized operations.\0";
        case DDERR_NOZBUFFERHW:
            return "Operation could not be carried out because there is no hardware support for zbuffer blitting.\0";
        case DDERR_NOZOVERLAYHW:
            return "Overlay surfaces could not be z layered based on their BltOrder because the hardware does not support z layering of overlays.\0";
        case DDERR_OUTOFCAPS:
            return "The hardware needed for the requested operation has already been allocated.\0";
        case DDERR_OUTOFMEMORY:
            return "DirectDraw does not have enough memory to perform the operation.\0";
        case DDERR_OUTOFVIDEOMEMORY:
            return "DirectDraw does not have enough memory to perform the operation.\0";
        case DDERR_OVERLAYCANTCLIP:
            return "The hardware does not support clipped overlays.\0";
        case DDERR_OVERLAYCOLORKEYONLYONEACTIVE:
            return "Can only have ony color key active at one time for overlays.\0";
        case DDERR_OVERLAYNOTVISIBLE:
            return "Returned when GetOverlayPosition is called on a hidden overlay.\0";
        case DDERR_PALETTEBUSY:
            return "Access to this palette is being refused because the palette is already locked by another thread.\0";
        case DDERR_PRIMARYSURFACEALREADYEXISTS:
            return "This process already has created a primary surface.\0";
        case DDERR_REGIONTOOSMALL:
            return "Region passed to Clipper::GetClipList is too small.\0";
        case DDERR_SURFACEALREADYATTACHED:
            return "This surface is already attached to the surface it is being attached to.\0";
        case DDERR_SURFACEALREADYDEPENDENT:
            return "This surface is already a dependency of the surface it is being made a dependency of.\0";
        case DDERR_SURFACEBUSY:
            return "Access to this surface is being refused because the surface is already locked by another thread.\0";
        case DDERR_SURFACEISOBSCURED:
            return "Access to surface refused because the surface is obscured.\0";
        case DDERR_SURFACELOST:
            return "Access to this surface is being refused because the surface memory is gone. The DirectDrawSurface object representing this surface should have Restore called on it.\0";
        case DDERR_SURFACENOTATTACHED:
            return "The requested surface is not attached.\0";
        case DDERR_TOOBIGHEIGHT:
            return "Height requested by DirectDraw is too large.\0";
        case DDERR_TOOBIGSIZE:
            return "Size requested by DirectDraw is too large, but the individual height and width are OK.\0";
        case DDERR_TOOBIGWIDTH:
            return "Width requested by DirectDraw is too large.\0";
        case DDERR_UNSUPPORTED:
            return "Action not supported.\0";
        case DDERR_UNSUPPORTEDFORMAT:
            return "FOURCC format requested is unsupported by DirectDraw.\0";
        case DDERR_UNSUPPORTEDMASK:
            return "Bitmask in the pixel format requested is unsupported by DirectDraw.\0";
        case DDERR_VERTICALBLANKINPROGRESS:
            return "Vertical blank is in progress.\0";
        case DDERR_WASSTILLDRAWING:
            return "Informs DirectDraw that the previous Blt which is transfering information to or from this Surface is incomplete.\0";
        case DDERR_WRONGMODE:
            return "This surface can not be restored because it was created in a different mode.\0";
        case DDERR_XALIGN:
            return "Rectangle provided was not horizontally aligned on required boundary.\0";
        default:
            return "Unrecognized error value.\0";
    }
}

bool Bitmap::DDErrH(HRESULT ddrval, int lineno)
{
	switch(ddrval)
	{
	case DD_OK:
		{
			return false;								// exit now
		}
	case DDERR_SURFACELOST:
		{
			if(ddsP->IsLost()!=DD_OK)
			{
				HRESULT ddrval;
				START_DDOP();
				ddrval=ddsP->Restore();
				END_DDOP(ddrval);
			}
			if(ddsB->IsLost()!=DD_OK)
			{
				HRESULT ddrval;
				START_DDOP();
				ddrval=ddsB->Restore();
				END_DDOP(ddrval);
			}
			return true;								// try again
		}
	case DDERR_WASSTILLDRAWING:
	case DDERR_SURFACEBUSY:
		{
			// Let something else run while DD is busy
			Sleep(0);
			return true;								// try again
		}
	default:
		{
			// Decode and throw
			DECLARE_ERROR(e, D3DAppErrorToString(ddrval));
			e.lineno=lineno;
			e.external=true;
			e.win32code=ddrval;
			throw e;
		}
	}
	return false;
}

Bitmap::Bitmap()
{
	memset(this, 0, sizeof(*this));
}
bool Bitmap::GDIOnly;

void Bitmap::Init()
{
	HRESULT ddrval;

    // create the main DirectDraw object
	START_DDOP();
    ddrval = DirectDrawCreate( NULL, &ddH, NULL );
	END_DDOP(ddrval);

	// Set to non-exclusive use
	START_DDOP();
    ddrval = ddH->SetCooperativeLevel( hwnd, DDSCL_NORMAL );
	END_DDOP(ddrval);

	// Create palette for the primary surface & attach
	if(bpp==8)
	{
		int i;
		TEST_NOMEM(paltemplate=(PALETTEENTRY *) calloc(256, sizeof(PALETTEENTRY)));

		// build a 332 palette as the default.
	    for (i=0; i<256; i++)
	    {
		    paltemplate[i].peRed   = (BYTE)(((i >> 5) & 0x07) * 255 / 7);
		    paltemplate[i].peGreen = (BYTE)(((i >> 2) & 0x07) * 255 / 7);
		    paltemplate[i].peBlue  = (BYTE)(((i >> 0) & 0x03) * 255 / 3);
		    paltemplate[i].peFlags = PC_NOCOLLAPSE;
	    }

		// castrate the palette according to windows ...
		// Set the first and last 10 entries to match the system palette.
		for(i=0;i<10;i++)
		{
		    paltemplate[i].peFlags  = paltemplate[i+246].peFlags = PC_EXPLICIT;
		    paltemplate[i].peRed = i;
		    paltemplate[i+246].peRed = i+246;
		}

	}

	// Create the screen refreshed event
	refreshevent=CreateEvent(NULL, FALSE, TRUE, NULL);

    // Create the primary surface, palette and clipper
	modeChanged();

	// Create cleared back buffer surface and attach palette
	ddsB=createBBSurface(width, height);
	bmpDC=NULL;

	// NT4 may not support the use of 8 bit surfaces in non 8 bit modes
	do
	{
		HDC dc;
		ddrval=ddsB->GetDC(&dc);
		if(ddrval!=DDERR_CANTCREATEDC)
		{
			ddsB->ReleaseDC(dc);
			NTmodeswitch=false;
			break;
		}
		if(MessageBox(NULL, "It has been detected that your system cannot execute Flow in non-256 colour modes.\nDo you want me to temporarily switch to a 256 colour mode?", "Question from Flow", MB_ICONQUESTION | MB_YESNO)==IDYES)
		{
			DDSURFACEDESC desc={sizeof(DDSURFACEDESC)};

			START_DDOP();
		    ddrval = ddH->GetDisplayMode(&desc);
			END_DDOP(ddrval);

			START_DDOP();
			ddrval = ddH->SetCooperativeLevel( hwnd, DDSCL_ALLOWREBOOT | 
				DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
			END_DDOP(ddrval);
			START_DDOP();
		    ddrval = ddH->SetDisplayMode(desc.dwWidth, desc.dwHeight, 8);
			END_DDOP(ddrval);
			START_DDOP();
			ddrval = ddH->SetCooperativeLevel( hwnd, DDSCL_NORMAL );
			END_DDOP(ddrval);

			PostMessage(hwnd, WM_DISPLAYCHANGE, 0, 0);
			NTmodeswitch=true;
		}
		else
			PostQuitMessage(0);
	} while(0);

}

LPDIRECTDRAWSURFACE Bitmap::createBBSurface(u32 w, u32 h)
{
	HRESULT ddrval;
	LPDIRECTDRAWSURFACE ddsB;
	// Create the back buffer surface
	START_DDOP();
    ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT;
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
    ddsd.dwWidth = w;
    ddsd.dwHeight = h;
	ddsd.ddpfPixelFormat.dwSize=sizeof(ddsd.ddpfPixelFormat);
	if(bpp==8)
		ddsd.ddpfPixelFormat.dwFlags=DDPF_RGB | DDPF_PALETTEINDEXED8;
	else
		ddsd.ddpfPixelFormat.dwFlags=DDPF_RGB;
	ddsd.ddpfPixelFormat.dwRGBBitCount=bpp;
    ddrval = ddH->CreateSurface( &ddsd, &ddsB, NULL );
	END_DDOP(ddrval);
	if(bpp==8)
	{
		// Set back buffer to use palette
		START_DDOP();
		ddrval=ddsB->SetPalette(ddpP);
		END_DDOP(ddrval);
	}
	// Clear the surface
	{
		DDBLTFX ddbltfx;

		START_DDOP();
		ddbltfx.dwSize = sizeof(ddbltfx);
		ddbltfx.dwFillColor = 0;
		ddrval = ddsB->Blt(NULL, NULL, NULL, DDBLT_COLORFILL, &ddbltfx);
		END_DDOP(ddrval);
	}
	return ddsB;
}

Bitmap::~Bitmap()
{
	if(ddH!=NULL)
	{
		releaseDC();
		if(fullscreen) setFullScreen(0);
		if(paltemplate)
		{
			free(paltemplate);
			paltemplate=NULL;
		}
		if(ddsP!=NULL)
		{
			if(ddcP!=NULL)
			{
				ddcP->Release();
				ddcP=NULL;
			}
			if(ddpP!=NULL)
			{
				ddpP->Release();
				ddpP=NULL;
			}
			ddsP->Release();
			ddsP=NULL;
		}
		if(ddsB!=NULL)
		{
			if(bmpDC!=NULL)
			{
				ddsB->ReleaseDC(bmpDC);
				bmpDC=0;
			}
			ddsB->Release();
			ddsB=NULL;
		}
		if(NTmodeswitch)
		{
			HRESULT ddrval;
			START_DDOP();
			ddrval = ddH->SetCooperativeLevel( hwnd, DDSCL_ALLOWREBOOT | 
				DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
			END_DDOP(ddrval);
			START_DDOP();
			ddrval = ddH->RestoreDisplayMode();
			END_DDOP(ddrval);
			START_DDOP();
			ddrval = ddH->SetCooperativeLevel( hwnd, DDSCL_NORMAL );
			END_DDOP(ddrval);
			NTmodeswitch=false;
		}
		ddH->Release();
		ddH=NULL;
	}
	if(refreshevent)
		CloseHandle(refreshevent);
}

void Bitmap::setSize(u32 w, u32 h)
{
	LPDIRECTDRAWSURFACE newddsB;
	HRESULT ddrval;
	RECT s;
	// Create a new back buffer and blit the old one into it
	newddsB=createBBSurface(w, h);
	START_DDOP();
	s.left=0;
	s.top=0;
	s.right=width;
	s.bottom=height;
	ddrval=newddsB->BltFast(0, 0, ddsB, &s, DDBLTFAST_NOCOLORKEY);
	END_DDOP(ddrval);
	ddsB->Release();
	ddsB=newddsB;
	width=w;
	height=h;
}

void Bitmap::paint(HDC dc, s32 x, s32 y, u32 scale)
{
	HRESULT ddrval=DDERR_UNSUPPORTED;
	RECT d,s;
	int count=0;

	try
	{
		START_DDOP();
		s.left=0;
		s.top=0;
		s.right=width;
		s.bottom=height;
		d.left=x;
		d.top=y;
		d.right=x+width*scale;
		d.bottom=y+height*scale;
		{	// Rather than spin the processor, we would prefer to give up our time slice
			WaitForSingleObject(refreshevent, INFINITE);
            ddH->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, NULL);
			timeSetEvent(refreshdelay, 0, (LPTIMECALLBACK) refreshevent, NULL, TIME_ONESHOT|TIME_CALLBACK_EVENT_SET);
		}
		if(!GDIOnly)
			ddrval=ddsP->Blt(&d, ddsB, &s, DDBLT_ASYNC, NULL);
		if(ddrval==DDERR_UNSUPPORTED && primarybpp!=8)
		{
			// Not capable of 8bpp->Xbpp blits then. Ok, do it via GDI then
			HDC tdc;
			POINT p={ d.left, d.top };
			ScreenToClient(hwnd, &p);
			tdc=getDC();
			if(scale==1)
			{
				TEST_WIN32ERR(!BitBlt(dc, p.x, p.y, d.right-d.left, d.bottom-d.top,
					tdc, 0, 0, SRCCOPY));
			}
			else
			{
				TEST_WIN32ERR(!StretchBlt(dc, p.x, p.y, d.right-d.left, d.bottom-d.top,
					tdc, s.left, s.top, s.right-s.left, s.bottom-s.top, SRCCOPY));
			}
			releaseDC();
			ddrval=DD_OK;
		}
		if(count++>1000) ddrval=DD_OK;
		END_DDOP(ddrval);
	}
	catch(Error e)
	{
		return;
	}
}

HDC Bitmap::getDC()
{
	if(bmpDC==NULL)
	{
		HRESULT ddrval;
		START_DDOP();
		ddrval=ddsB->GetDC(&bmpDC);
		END_DDOP(ddrval);
	}
	return bmpDC;
}

void Bitmap::releaseDC()
{
	if(bmpDC!=NULL)
	{
		HRESULT ddrval;
		START_DDOP();
		ddrval=ddsB->ReleaseDC(bmpDC);
		END_DDOP(ddrval);
		bmpDC=NULL;
	}
}

void *Bitmap::getAddr()
{
	if(bmpaddr==NULL)
	{
		HRESULT ddrval;
		RECT r;
		START_DDOP();
		r.left=0;
		r.top=0;
		r.right=width;
		r.bottom=height;
		ddrval=ddsB->Lock(&r, &ddsd, 0, NULL);
		END_DDOP(ddrval);
		bmpaddr=ddsd.lpSurface;
	}
	return bmpaddr;
}

void Bitmap::releaseAddr()
{
	if(bmpaddr!=NULL)
	{
		HRESULT ddrval;
		START_DDOP();
		ddrval=ddsB->Unlock(bmpaddr);
		END_DDOP(ddrval);
		bmpaddr=NULL;
	}
}

void Bitmap::modeChanged()
{
	HRESULT ddrval;
	// Detect the current screen refresh rate
	{
		u32 vals[10], t;
        ddH->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, NULL);
		for(int n=0; n<10; n++)
		{
			t=timeGetTime();
            ddH->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, NULL);
			vals[n]=timeGetTime()-t;
		}
		refreshdelay=0;
		for(int n=0; n<10; n++)
			refreshdelay+=vals[n];
		refreshdelay/=10;
		// We may have been unlucky and have some wildly high number in there
		for(int n=0; n<10; n++)
			if(vals[n]>=refreshdelay*2-1) vals[n]=refreshdelay;
		refreshdelay=0;
		for(int n=0; n<10; n++)
			refreshdelay+=vals[n];
		refreshdelay/=10;
		// It should be around 12 to 17ms
	}

	// Recreate the primary surface & palette

	// Determine current primary bpp
	{
		HDC dc;
		int w,h;
		char msg[256]="";
		dc=CreateCompatibleDC(NULL);
		w=GetDeviceCaps(dc, HORZRES);
		h=GetDeviceCaps(dc, VERTRES);
		primarybpp=GetDeviceCaps(dc, BITSPIXEL);
		DeleteDC(dc);
	}
	if(ddpP)
	{
		ddpP->Release();
		ddpP=NULL;
	}
	if(paltemplate)
	{
		START_DDOP();
		ddrval=ddH->CreatePalette(DDPCAPS_8BIT, paltemplate, &ddpP, NULL);
		END_DDOP(ddrval);
	}
	else
		ddpP=NULL;

	if(ddsP)
	{
		ddsP->Release();
		ddsP=NULL;
	}
	START_DDOP();
    ddsd.dwSize = sizeof( ddsd );
    ddsd.dwFlags = DDSD_CAPS;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
    ddrval = ddH->CreateSurface( &ddsd, &ddsP, NULL );
	END_DDOP(ddrval);

	// Create the clipper for the primary surface & attach to our window
	if(ddcP)
	{
		ddcP->Release();
		ddcP=NULL;
	}
	START_DDOP();
	ddrval=ddH->CreateClipper(0, &ddcP, NULL);
	END_DDOP(ddrval);
	START_DDOP();
	ddrval=ddcP->SetHWnd(0, hwnd);
	END_DDOP(ddrval);
	START_DDOP();
	ddrval=ddsP->SetClipper(ddcP);
	END_DDOP(ddrval);

	// Attach palette to primary surface
	installPalette();
}

void Bitmap::installPalette()
{
	if(bpp==8 && ddpP && primarybpp==8)
	{
		HRESULT ddrval;
		try
		{
			START_DDOP();
			ddrval=ddsP->SetPalette(ddpP);
			END_DDOP(ddrval);
		}
		catch(Error e)
		{
			return;
		}
	}
}

void Bitmap::setFullScreen(bool setting)
{
	static HMENU wndmenu;
	static LONG wndstyle, wndstyleex;
	static RECT wndpos;
	HRESULT ddrval;
	if(setting==fullscreen) return;
	if(setting)
	{
		// Request exclusive access & set video mode
		GetWindowRect(hwnd, &wndpos);
		wndstyle=SetWindowLong(hwnd, GWL_STYLE, 0);
		wndstyleex=SetWindowLong(hwnd, GWL_EXSTYLE, 0);
		wndmenu=GetMenu(hwnd);
		SetMenu(hwnd, NULL);
		START_DDOP();
		ddrval = ddH->SetCooperativeLevel( hwnd, DDSCL_ALLOWREBOOT | 
			DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
		END_DDOP(ddrval);
		START_DDOP();
		ddrval = ddH->SetDisplayMode( width, height, bpp);
		END_DDOP(ddrval);
		fullscreen=true;
	}
	else
	{
		// Restore original mode and return to normal functioning
		START_DDOP();
		ddrval = ddH->RestoreDisplayMode();
		END_DDOP(ddrval);
		START_DDOP();
		ddrval = ddH->SetCooperativeLevel( hwnd, DDSCL_NORMAL );
		END_DDOP(ddrval);
		SetWindowLong(hwnd, GWL_STYLE, wndstyle);
		SetWindowLong(hwnd, GWL_EXSTYLE, wndstyleex);
		SetMenu(hwnd, wndmenu);
		SetWindowPos(hwnd, NULL, wndpos.left, wndpos.top, 0, 0,
			SWP_NOZORDER|SWP_NOSIZE|SWP_NOACTIVATE);
		fullscreen=false;
	}
	// Inform window palette and mode has changed
	SendMessage(hwnd, WM_DISPLAYCHANGE, bpp, (height<<16) | width);
}

u32 Bitmap::RGBtoPhysical(u32 rgb)
{
    COLORREF rgbT;
    HDC hdc;
    DWORD dw = CLR_INVALID;
	HPEN drawingpen;

#if 0
	if(bpp!=8)
	{
		// Is a 332 (rgb) palette, hence ...
		int r,g,b;
		r=(rgb>>0) & 0xff;
		g=(rgb>>8) & 0xff;
		b=(rgb>>16) & 0xff;
		r>>=5;
		g>>=5;
		b>>=6;
		return (r<<5) | (g<<2) | (b<<0);
	}
#endif

    //
    //  use GDI SetPixel to color match for us
    //
	drawingpen=CreatePen(PS_SOLID, 1, rgb);
	hdc=getDC();
    {
        rgbT = GetPixel(hdc, 0, 0);             // save current pixel value
		SelectObject(hdc, drawingpen);
		MoveToEx(hdc, 0, 0, NULL);
		LineTo(hdc, 1, 0);
        releaseDC();
    }
	DeleteObject(drawingpen);

    //
    // now lock the surface so we can read back the converted color
    //
    ddsd.dwSize = sizeof(ddsd);


    {
        dw  = *(DWORD *)getAddr();
        dw &= (1 << ddsd.ddpfPixelFormat.dwRGBBitCount)-1;  // mask it to bpp
        releaseAddr();
    }

    //
    //  now put the color that was there back.
    //
	hdc=getDC();
    if (hdc!=NULL)
    {
        SetPixel(hdc, 0, 0, rgbT);
        releaseDC();
    }

    return dw;
}

#pragma pack(push,1)
struct FileHeader
{
	u32 magic2;
	u32 version;
	u32 width;
	u32 height;
	u32 bpp;
	u32 frames;
};

struct Frame
{
	u8 rptc;
	union
	{
		u8 col8;
		u16 col16;
		u32 col32;
	};
};
#pragma pack(pop)

void Bitmap::load(HANDLE inh)
{
	Bitmap *bmp;
	FileHeader fh;
	DWORD r;
	Frame fl;
	u32 flcount;
	int framesize;

	TEST_WIN32ERR(!ReadFile(inh, &fh, sizeof(fh), &r, NULL));
	if(fh.magic2!=*(u32 *) "BMAP")
	{
		DECLARE_ERROR(e, "This file does not contain Flow bitmap information");
		throw e;
	}
	if(fh.version!=90)
	{
		DECLARE_ERROR(e, "This file is in an incompatible version");
		throw e;
	}
	// Create a new bitmap object to hold the file
	framesize=sizeof(u8);
	if(fh.bpp==8)
	{
		framesize+=sizeof(u8);
		bmp=new ScreenImage8(hwnd, fh.width, fh.height);
	}
	else if(fh.bpp==16)
	{
		framesize+=sizeof(u16);
		bmp=new ScreenImage16(hwnd, fh.width, fh.height);
	}
	else if(fh.bpp==32)
	{
		framesize+=sizeof(u32);
		bmp=new ScreenImage32(hwnd, fh.width, fh.height);
	}
	else
	{
		DECLARE_ERROR(e, "Unknown image bpp in input file");
		throw e;
	}

	u32 offset=0;
	bmp->getAddr();
	for(flcount=0; flcount<fh.frames; flcount++)
	{
		int rptc;
		TEST_WIN32ERR(!ReadFile(inh, &fl, framesize, &r, NULL));
		for(rptc=0; rptc<fl.rptc; rptc++)
		{
			if(fh.bpp==8)
				bmp->writepixel(offset++, fl.col8);
			else if(fh.bpp==16)
				bmp->writepixel(offset++, fl.col16);
			else if(fh.bpp==32)
				bmp->writepixel(offset++, fl.col32);
			if(offset>fh.width*fh.height)
			{
				DECLARE_ERROR(e, "Corrupted or invalid image data");
				throw e;
			}
		}
	}
	bmp->releaseAddr();

	// Ok, bmp is now new bitmap
	// Set ourselves to be it
	void *tbmp;
	int size;
	size=sizeof(*bmp);
	TEST_NOMEM(tbmp=malloc(size));
	memcpy(tbmp, this, size);
	memcpy(this, bmp, size);
	memcpy(bmp, tbmp, size);
	free(tbmp);
	delete bmp;
}

void Bitmap::save(HANDLE outh)
{
#ifdef SHAREWARE
	DECLARE_ERROR(e, "Not implemented in the shareware build");
	throw e;
#else
	FileHeader fh;
	DWORD w;
	Frame *fl, *flptr;
	int flcount,fllen, framesize;

	// Build header & create frame store
	fh.magic2=*(u32 *) "BMAP";
	fh.version=90;
	fh.width=width;
	fh.height=height;
	fh.bpp=bpp;

	framesize=sizeof(u8);
	if(fh.bpp==8)
		framesize+=sizeof(u8);
	else if(fh.bpp==16)
		framesize+=sizeof(u16);
	else if(fh.bpp==32)
		framesize+=sizeof(u32);
	else
	{
		DECLARE_ERROR(e, "Unknown bpp");
		throw e;
	}
	TEST_NOMEM(fl=(Frame *) calloc(256, framesize));
	flptr=fl;
	flcount=0;
	fllen=256;

	// Run along the screen, counting repetitions of colour
	u32 offset=0;
	do
	{
		u32 currcol;
		if(fh.bpp==8)
			flptr->col8=u8(currcol=readpixel(offset));
		else if(fh.bpp==16)
			flptr->col16=u16(currcol=readpixel(offset));
		else if(fh.bpp==32)
			flptr->col32=u32(currcol=readpixel(offset));
		flptr->rptc=0;
		do
		{
			flptr->rptc++;
			offset++;
		}
		while(readpixel(offset)==currcol && flptr->rptc<255
			&& offset<fh.width*fh.height);
		flptr=(Frame *)(u32(flptr)+framesize);
		if(++flcount>=fllen)
		{
			Frame *nfl;
			flptr=(Frame *)(u32(flptr)-u32(fl));
			TEST_NOMEM(nfl=(Frame *) realloc(fl, (256+fllen)*framesize));
			fl=nfl;
			flptr=(Frame *)(u32(flptr)+u32(fl));
			memset(flptr, 0, 256*framesize);
			fllen+=256;
		}
	} while(offset<fh.width*fh.height);
	fh.frames=flcount;
	TEST_WIN32ERR(!WriteFile(outh, &fh, sizeof(FileHeader), &w, NULL));
	TEST_WIN32ERR(!WriteFile(outh, fl, fh.frames*framesize, &w, NULL));
#endif
}
